@@ -1,15 +1,37 @@
 /*	$OpenBSD: shf.c,v 1.15 2006/04/02 00:48:33 deraadt Exp $	*/
 
+/*-
+ * Copyright (c) 2003, 2004, 2005, 2006, 2007, 2008, 2009
+ *	Thorsten Glaser <tg@mirbsd.org>
+ *
+ * Provided that these terms and disclaimer and all copyright notices
+ * are retained or reproduced in an accompanying document, permission
+ * is granted to deal in this work without restriction, including un-
+ * limited rights to use, publicly perform, distribute, sell, modify,
+ * merge, give away, or sublicence.
+ *
+ * This work is provided "AS IS" and WITHOUT WARRANTY of any kind, to
+ * the utmost extent permitted by applicable law, neither express nor
+ * implied; without malicious intent or gross negligence. In no event
+ * may a licensor, author or contributor be held liable for indirect,
+ * direct, other damage, loss, or other issues arising in any way out
+ * of dealing in the work, even if advised of the possibility of such
+ * damage or existence of a defect, except proven that it results out
+ * of said person's immediate fault when using the work as intended.
+ *-
+ * Use %lX instead of %p and floating point isn't supported at all.
+ */
+
 #include "sh.h"
 
-__RCSID("$MirOS: src/bin/mksh/shf.c,v 1.21 2008/05/17 18:47:02 tg Exp $");
+__RCSID("$MirOS: src/bin/mksh/shf.c,v 1.40 2010/09/14 21:26:17 tg Exp $");
 
 /* flags to shf_emptybuf() */
 #define EB_READSW	0x01	/* about to switch to reading */
 #define EB_GROW		0x02	/* grow buffer if necessary (STRING+DYNAMIC) */
 
 /*
- * Replacement stdio routines.  Stdio is too flakey on too many machines
+ * Replacement stdio routines. Stdio is too flakey on too many machines
  * to be useful when you have multiple processes using the same underlying
  * file descriptors.
  */
@@ -17,19 +39,20 @@ __RCSID("$MirOS: src/bin/mksh/shf.c,v 1.21 2008/05/17 18:47:02 tg Exp $");
 static int shf_fillbuf(struct shf *);
 static int shf_emptybuf(struct shf *, int);
 
-/* Open a file.  First three args are for open(), last arg is flags for
- * this package.  Returns NULL if file could not be opened, or if a dup
+/* Open a file. First three args are for open(), last arg is flags for
+ * this package. Returns NULL if file could not be opened, or if a dup
  * fails.
  */
 struct shf *
 shf_open(const char *name, int oflags, int mode, int sflags)
 {
 	struct shf *shf;
-	int bsize = sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
+	int bsize = /* at most 512 */
+	    sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
 	int fd;
 
 	/* Done before open so if alloca fails, fd won't be lost. */
-	shf = (struct shf *)alloc(sizeof (struct shf) + bsize, ATEMP);
+	shf = alloc(sizeof(struct shf) + bsize, ATEMP);
 	shf->areap = ATEMP;
 	shf->buf = (unsigned char *)&shf[1];
 	shf->bsize = bsize;
@@ -39,7 +62,7 @@ shf_open(const char *name, int oflags, int mode, int sflags)
 	fd = open(name, oflags, mode);
 	if (fd < 0) {
 		afree(shf, shf->areap);
-		return NULL;
+		return (NULL);
 	}
 	if ((sflags & SHF_MAPHI) && fd < FDBASE) {
 		int nfd;
@@ -48,7 +71,7 @@ shf_open(const char *name, int oflags, int mode, int sflags)
 		close(fd);
 		if (nfd < 0) {
 			afree(shf, shf->areap);
-			return NULL;
+			return (NULL);
 		}
 		fd = nfd;
 	}
@@ -56,14 +79,15 @@ shf_open(const char *name, int oflags, int mode, int sflags)
 	sflags |= (oflags & O_ACCMODE) == O_RDONLY ? SHF_RD :
 	    ((oflags & O_ACCMODE) == O_WRONLY ? SHF_WR : SHF_RDWR);
 
-	return shf_reopen(fd, sflags, shf);
+	return (shf_reopen(fd, sflags, shf));
 }
 
-/* Set up the shf structure for a file descriptor.  Doesn't fail. */
+/* Set up the shf structure for a file descriptor. Doesn't fail. */
 struct shf *
 shf_fdopen(int fd, int sflags, struct shf *shf)
 {
-	int bsize = sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
+	int bsize = /* at most 512 */
+	    sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
 
 	/* use fcntl() to figure out correct read/write flags */
 	if (sflags & SHF_GETFL) {
@@ -88,16 +112,16 @@ shf_fdopen(int fd, int sflags, struct shf *shf)
 	}
 
 	if (!(sflags & (SHF_RD | SHF_WR)))
-		internal_errorf("shf_fdopen: missing read/write");
+		internal_errorf("%s: %s", "shf_fdopen", "missing read/write");
 
 	if (shf) {
 		if (bsize) {
-			shf->buf = (unsigned char *)alloc(bsize, ATEMP);
+			shf->buf = alloc(bsize, ATEMP);
 			sflags |= SHF_ALLOCB;
 		} else
 			shf->buf = NULL;
 	} else {
-		shf = (struct shf *)alloc(sizeof (struct shf) + bsize, ATEMP);
+		shf = alloc(sizeof(struct shf) + bsize, ATEMP);
 		shf->buf = (unsigned char *)&shf[1];
 		sflags |= SHF_ALLOCS;
 	}
@@ -113,14 +137,15 @@ shf_fdopen(int fd, int sflags, struct shf *shf)
 	shf->bsize = bsize;
 	if (sflags & SHF_CLEXEC)
 		fcntl(fd, F_SETFD, FD_CLOEXEC);
-	return shf;
+	return (shf);
 }
 
 /* Set up an existing shf (and buffer) to use the given fd */
 struct shf *
 shf_reopen(int fd, int sflags, struct shf *shf)
 {
-	int bsize = sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
+	int bsize = /* at most 512 */
+	    sflags & SHF_UNBUF ? (sflags & SHF_RD ? 1 : 0) : SHF_BSIZE;
 
 	/* use fcntl() to figure out correct read/write flags */
 	if (sflags & SHF_GETFL) {
@@ -145,9 +170,9 @@ shf_reopen(int fd, int sflags, struct shf *shf)
 	}
 
 	if (!(sflags & (SHF_RD | SHF_WR)))
-		internal_errorf("shf_reopen: missing read/write");
+		internal_errorf("%s: %s", "shf_reopen", "missing read/write");
 	if (!shf || !shf->buf || shf->bsize < bsize)
-		internal_errorf("shf_reopen: bad shf/buf/bsize");
+		internal_errorf("%s: %s", "shf_reopen", "bad shf/buf/bsize");
 
 	/* assumes shf->buf and shf->bsize already set up */
 	shf->fd = fd;
@@ -160,27 +185,26 @@ shf_reopen(int fd, int sflags, struct shf *shf)
 	shf->errno_ = 0;
 	if (sflags & SHF_CLEXEC)
 		fcntl(fd, F_SETFD, FD_CLOEXEC);
-	return shf;
+	return (shf);
 }
 
-/* Open a string for reading or writing.  If reading, bsize is the number
- * of bytes that can be read.  If writing, bsize is the maximum number of
- * bytes that can be written.  If shf is not null, it is filled in and
- * returned, if it is null, shf is allocated.  If writing and buf is null
+/* Open a string for reading or writing. If reading, bsize is the number
+ * of bytes that can be read. If writing, bsize is the maximum number of
+ * bytes that can be written. If shf is not null, it is filled in and
+ * returned, if it is null, shf is allocated. If writing and buf is null
  * and SHF_DYNAMIC is set, the buffer is allocated (if bsize > 0, it is
- * used for the initial size).  Doesn't fail.
+ * used for the initial size). Doesn't fail.
  * When writing, a byte is reserved for a trailing null - see shf_sclose().
  */
 struct shf *
 shf_sopen(char *buf, int bsize, int sflags, struct shf *shf)
 {
 	/* can't have a read+write string */
-	if (!(sflags & (SHF_RD | SHF_WR)) ||
-	    (sflags & (SHF_RD | SHF_WR)) == (SHF_RD | SHF_WR))
-		internal_errorf("shf_sopen: flags 0x%x", sflags);
+	if (!(!(sflags & SHF_RD) ^ !(sflags & SHF_WR)))
+		internal_errorf("%s: flags 0x%X", "shf_sopen", sflags);
 
 	if (!shf) {
-		shf = (struct shf *)alloc(sizeof (struct shf), ATEMP);
+		shf = alloc(sizeof(struct shf), ATEMP);
 		sflags |= SHF_ALLOCS;
 	}
 	shf->areap = ATEMP;
@@ -200,7 +224,7 @@ shf_sopen(char *buf, int bsize, int sflags, struct shf *shf)
 	shf->errno_ = 0;
 	shf->bsize = bsize;
 
-	return shf;
+	return (shf);
 }
 
 /* Flush and close file descriptor, free the shf structure */
@@ -219,7 +243,7 @@ shf_close(struct shf *shf)
 	else if (shf->flags & SHF_ALLOCB)
 		afree(shf->buf, shf->areap);
 
-	return ret;
+	return (ret);
 }
 
 /* Flush and close file descriptor, don't free file structure */
@@ -238,7 +262,7 @@ shf_fdclose(struct shf *shf)
 		shf->fd = -1;
 	}
 
-	return ret;
+	return (ret);
 }
 
 /* Close a string - if it was opened for writing, it is null terminated;
@@ -257,42 +281,42 @@ shf_sclose(struct shf *shf)
 	}
 	if (shf->flags & SHF_ALLOCS)
 		afree(shf, shf->areap);
-	return (char *) s;
+	return ((char *)s);
 }
 
 /* Un-read what has been read but not examined, or write what has been
- * buffered.  Returns 0 for success, EOF for (write) error.
+ * buffered. Returns 0 for success, EOF for (write) error.
  */
 int
 shf_flush(struct shf *shf)
 {
 	if (shf->flags & SHF_STRING)
-		return (shf->flags & SHF_WR) ? EOF : 0;
+		return ((shf->flags & SHF_WR) ? EOF : 0);
 
 	if (shf->fd < 0)
-		internal_errorf("shf_flush: no fd");
+		internal_errorf("%s: %s", "shf_flush", "no fd");
 
 	if (shf->flags & SHF_ERROR) {
 		errno = shf->errno_;
-		return EOF;
+		return (EOF);
 	}
 
 	if (shf->flags & SHF_READING) {
 		shf->flags &= ~(SHF_EOF | SHF_READING);
 		if (shf->rnleft > 0) {
-			lseek(shf->fd, (off_t) -shf->rnleft, SEEK_CUR);
+			lseek(shf->fd, (off_t)-shf->rnleft, SEEK_CUR);
 			shf->rnleft = 0;
 			shf->rp = shf->buf;
 		}
-		return 0;
+		return (0);
 	} else if (shf->flags & SHF_WRITING)
-		return shf_emptybuf(shf, 0);
+		return (shf_emptybuf(shf, 0));
 
-	return 0;
+	return (0);
 }
 
-/* Write out any buffered data.  If currently reading, flushes the read
- * buffer.  Returns 0 for success, EOF for (write) error.
+/* Write out any buffered data. If currently reading, flushes the read
+ * buffer. Returns 0 for success, EOF for (write) error.
  */
 static int
 shf_emptybuf(struct shf *shf, int flags)
@@ -300,16 +324,16 @@ shf_emptybuf(struct shf *shf, int flags)
 	int ret = 0;
 
 	if (!(shf->flags & SHF_STRING) && shf->fd < 0)
-		internal_errorf("shf_emptybuf: no fd");
+		internal_errorf("%s: %s", "shf_emptybuf", "no fd");
 
 	if (shf->flags & SHF_ERROR) {
 		errno = shf->errno_;
-		return EOF;
+		return (EOF);
 	}
 
 	if (shf->flags & SHF_READING) {
 		if (flags & EB_READSW) /* doesn't happen */
-			return 0;
+			return (0);
 		ret = shf_flush(shf);
 		shf->flags &= ~SHF_READING;
 	}
@@ -321,10 +345,9 @@ shf_emptybuf(struct shf *shf, int flags)
 		 */
 		if (!(flags & EB_GROW) || !(shf->flags & SHF_DYNAMIC) ||
 		    !(shf->flags & SHF_ALLOCB))
-			return EOF;
+			return (EOF);
 		/* allocate more space for buffer */
-		nbuf = (unsigned char *)aresize(shf->buf, shf->wbsize * 2,
-		    shf->areap);
+		nbuf = aresize2(shf->buf, 2, shf->wbsize, shf->areap);
 		shf->rp = nbuf + (shf->rp - shf->buf);
 		shf->wp = nbuf + (shf->wp - shf->buf);
 		shf->rbsize += shf->wbsize;
@@ -353,7 +376,7 @@ shf_emptybuf(struct shf *shf, int flags)
 						    ntowrite);
 						shf->wp = shf->buf + ntowrite;
 					}
-					return EOF;
+					return (EOF);
 				}
 				buf += n;
 				ntowrite -= n;
@@ -362,7 +385,7 @@ shf_emptybuf(struct shf *shf, int flags)
 				shf->wp = shf->buf;
 				shf->wnleft = 0;
 				shf->flags &= ~SHF_WRITING;
-				return 0;
+				return (0);
 			}
 		}
 		shf->wp = shf->buf;
@@ -370,53 +393,52 @@ shf_emptybuf(struct shf *shf, int flags)
 	}
 	shf->flags |= SHF_WRITING;
 
-	return ret;
+	return (ret);
 }
 
-/* Fill up a read buffer.  Returns EOF for a read error, 0 otherwise. */
+/* Fill up a read buffer. Returns EOF for a read error, 0 otherwise. */
 static int
 shf_fillbuf(struct shf *shf)
 {
+	ssize_t n;
+
 	if (shf->flags & SHF_STRING)
-		return 0;
+		return (0);
 
 	if (shf->fd < 0)
-		internal_errorf("shf_fillbuf: no fd");
+		internal_errorf("%s: %s", "shf_fillbuf", "no fd");
 
 	if (shf->flags & (SHF_EOF | SHF_ERROR)) {
 		if (shf->flags & SHF_ERROR)
 			errno = shf->errno_;
-		return EOF;
+		return (EOF);
 	}
 
 	if ((shf->flags & SHF_WRITING) && shf_emptybuf(shf, EB_READSW) == EOF)
-		return EOF;
+		return (EOF);
 
 	shf->flags |= SHF_READING;
 
 	shf->rp = shf->buf;
 	while (1) {
-		shf->rnleft = blocking_read(shf->fd, (char *) shf->buf,
-		    shf->rbsize);
-		if (shf->rnleft < 0 && errno == EINTR &&
-		    !(shf->flags & SHF_INTERRUPT))
+		n = blocking_read(shf->fd, (char *)shf->buf, shf->rbsize);
+		if (n < 0 && errno == EINTR && !(shf->flags & SHF_INTERRUPT))
 			continue;
 		break;
 	}
-	if (shf->rnleft <= 0) {
-		if (shf->rnleft < 0) {
-			shf->flags |= SHF_ERROR;
-			shf->errno_ = errno;
-			shf->rnleft = 0;
-			shf->rp = shf->buf;
-			return EOF;
-		}
-		shf->flags |= SHF_EOF;
+	if (n < 0) {
+		shf->flags |= SHF_ERROR;
+		shf->errno_ = errno;
+		shf->rnleft = 0;
+		shf->rp = shf->buf;
+		return (EOF);
 	}
-	return 0;
+	if ((shf->rnleft = n) == 0)
+		shf->flags |= SHF_EOF;
+	return (0);
 }
 
-/* Read a buffer from shf.  Returns the number of bytes read into buf,
+/* Read a buffer from shf. Returns the number of bytes read into buf,
  * if no bytes were read, returns 0 if end of file was seen, EOF if
  * a read error occurred.
  */
@@ -427,10 +449,10 @@ shf_read(char *buf, int bsize, struct shf *shf)
 	int ncopy;
 
 	if (!(shf->flags & SHF_RD))
-		internal_errorf("shf_read: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_read", shf->flags);
 
 	if (bsize <= 0)
-		internal_errorf("shf_read: bsize %d", bsize);
+		internal_errorf("%s: %s %d", "shf_write", "bsize", bsize);
 
 	while (bsize > 0) {
 		if (shf->rnleft == 0 &&
@@ -446,12 +468,12 @@ shf_read(char *buf, int bsize, struct shf *shf)
 		shf->rnleft -= ncopy;
 	}
 	/* Note: fread(3S) returns 0 for errors - this doesn't */
-	return orig_bsize == bsize ? (shf_error(shf) ? EOF : 0) :
-	    orig_bsize - bsize;
+	return (orig_bsize == bsize ? (shf_error(shf) ? EOF : 0) :
+	    orig_bsize - bsize);
 }
 
-/* Read up to a newline or EOF.  The newline is put in buf; buf is always
- * null terminated.  Returns NULL on read error or if nothing was read before
+/* Read up to a newline or EOF. The newline is put in buf; buf is always
+ * null terminated. Returns NULL on read error or if nothing was read before
  * end of file, returns a pointer to the null byte in buf otherwise.
  */
 char *
@@ -462,19 +484,19 @@ shf_getse(char *buf, int bsize, struct shf *shf)
 	char *orig_buf = buf;
 
 	if (!(shf->flags & SHF_RD))
-		internal_errorf("shf_getse: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_getse", shf->flags);
 
 	if (bsize <= 0)
-		return NULL;
+		return (NULL);
 
 	--bsize;	/* save room for null */
 	do {
 		if (shf->rnleft == 0) {
 			if (shf_fillbuf(shf) == EOF)
-				return NULL;
+				return (NULL);
 			if (shf->rnleft == 0) {
 				*buf = '\0';
-				return buf == orig_buf ? NULL : buf;
+				return (buf == orig_buf ? NULL : buf);
 			}
 		}
 		end = (unsigned char *)memchr((char *) shf->rp, '\n',
@@ -489,37 +511,37 @@ shf_getse(char *buf, int bsize, struct shf *shf)
 		bsize -= ncopy;
 	} while (!end && bsize);
 	*buf = '\0';
-	return buf;
+	return (buf);
 }
 
-/* Returns the char read.  Returns EOF for error and end of file. */
+/* Returns the char read. Returns EOF for error and end of file. */
 int
 shf_getchar(struct shf *shf)
 {
 	if (!(shf->flags & SHF_RD))
-		internal_errorf("shf_getchar: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_getchar", shf->flags);
 
 	if (shf->rnleft == 0 && (shf_fillbuf(shf) == EOF || shf->rnleft == 0))
-		return EOF;
+		return (EOF);
 	--shf->rnleft;
-	return *shf->rp++;
+	return (*shf->rp++);
 }
 
-/* Put a character back in the input stream.  Returns the character if
+/* Put a character back in the input stream. Returns the character if
  * successful, EOF if there is no room.
  */
 int
 shf_ungetc(int c, struct shf *shf)
 {
 	if (!(shf->flags & SHF_RD))
-		internal_errorf("shf_ungetc: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_ungetc", shf->flags);
 
 	if ((shf->flags & SHF_ERROR) || c == EOF ||
 	    (shf->rp == shf->buf && shf->rnleft))
-		return EOF;
+		return (EOF);
 
 	if ((shf->flags & SHF_WRITING) && shf_emptybuf(shf, EB_READSW) == EOF)
-		return EOF;
+		return (EOF);
 
 	if (shf->rp == shf->buf)
 		shf->rp = shf->buf + shf->rbsize;
@@ -528,39 +550,39 @@ shf_ungetc(int c, struct shf *shf)
 		 * don't want to modify a string.
 		 */
 		if (shf->rp[-1] != c)
-			return EOF;
+			return (EOF);
 		shf->flags &= ~SHF_EOF;
 		shf->rp--;
 		shf->rnleft++;
-		return c;
+		return (c);
 	}
 	shf->flags &= ~SHF_EOF;
 	*--(shf->rp) = c;
 	shf->rnleft++;
-	return c;
+	return (c);
 }
 
-/* Write a character.  Returns the character if successful, EOF if
+/* Write a character. Returns the character if successful, EOF if
  * the char could not be written.
  */
 int
 shf_putchar(int c, struct shf *shf)
 {
 	if (!(shf->flags & SHF_WR))
-		internal_errorf("shf_putchar: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_putchar", shf->flags);
 
 	if (c == EOF)
-		return EOF;
+		return (EOF);
 
 	if (shf->flags & SHF_UNBUF) {
 		unsigned char cc = (unsigned char)c;
 		int n;
 
 		if (shf->fd < 0)
-			internal_errorf("shf_putchar: no fd");
+			internal_errorf("%s: %s", "shf_putchar", "no fd");
 		if (shf->flags & SHF_ERROR) {
 			errno = shf->errno_;
-			return EOF;
+			return (EOF);
 		}
 		while ((n = write(shf->fd, &cc, 1)) != 1)
 			if (n < 0) {
@@ -569,42 +591,42 @@ shf_putchar(int c, struct shf *shf)
 					continue;
 				shf->flags |= SHF_ERROR;
 				shf->errno_ = errno;
-				return EOF;
+				return (EOF);
 			}
 	} else {
 		/* Flush deals with strings and sticky errors */
 		if (shf->wnleft == 0 && shf_emptybuf(shf, EB_GROW) == EOF)
-			return EOF;
+			return (EOF);
 		shf->wnleft--;
 		*shf->wp++ = c;
 	}
 
-	return c;
+	return (c);
 }
 
-/* Write a string.  Returns the length of the string if successful, EOF if
+/* Write a string. Returns the length of the string if successful, EOF if
  * the string could not be written.
  */
 int
 shf_puts(const char *s, struct shf *shf)
 {
 	if (!s)
-		return EOF;
+		return (EOF);
 
-	return shf_write(s, strlen(s), shf);
+	return (shf_write(s, strlen(s), shf));
 }
 
-/* Write a buffer.  Returns nbytes if successful, EOF if there is an error. */
+/* Write a buffer. Returns nbytes if successful, EOF if there is an error. */
 int
 shf_write(const char *buf, int nbytes, struct shf *shf)
 {
 	int n, ncopy, orig_nbytes = nbytes;
 
 	if (!(shf->flags & SHF_WR))
-		internal_errorf("shf_write: flags %x", shf->flags);
+		internal_errorf("%s: flags 0x%X", "shf_write", shf->flags);
 
 	if (nbytes < 0)
-		internal_errorf("shf_write: nbytes %d", nbytes);
+		internal_errorf("%s: %s %d", "shf_write", "nbytes", nbytes);
 
 	/* Don't buffer if buffer is empty and we're writting a large amount. */
 	if ((ncopy = shf->wnleft) &&
@@ -618,39 +640,52 @@ shf_write(const char *buf, int nbytes, struct shf *shf)
 		shf->wnleft -= ncopy;
 	}
 	if (nbytes > 0) {
-		/* Flush deals with strings and sticky errors */
-		if (shf_emptybuf(shf, EB_GROW) == EOF)
-			return EOF;
-		if (nbytes > shf->wbsize) {
-			ncopy = nbytes;
-			if (shf->wbsize)
-				ncopy -= nbytes % shf->wbsize;
-			nbytes -= ncopy;
-			while (ncopy > 0) {
-				n = write(shf->fd, buf, ncopy);
-				if (n < 0) {
-					if (errno == EINTR &&
-					    !(shf->flags & SHF_INTERRUPT))
-						continue;
-					shf->flags |= SHF_ERROR;
-					shf->errno_ = errno;
-					shf->wnleft = 0;
-					/* Note: fwrite(3S) returns 0 for
-					 * errors - this doesn't */
-					return EOF;
+		if (shf->flags & SHF_STRING) {
+			/* resize buffer until there's enough space left */
+			while (nbytes > shf->wnleft)
+				if (shf_emptybuf(shf, EB_GROW) == EOF)
+					return (EOF);
+			/* then write everything into the buffer */
+		} else {
+			/* flush deals with sticky errors */
+			if (shf_emptybuf(shf, EB_GROW) == EOF)
+				return (EOF);
+			/* write chunks larger than window size directly */
+			if (nbytes > shf->wbsize) {
+				ncopy = nbytes;
+				if (shf->wbsize)
+					ncopy -= nbytes % shf->wbsize;
+				nbytes -= ncopy;
+				while (ncopy > 0) {
+					n = write(shf->fd, buf, ncopy);
+					if (n < 0) {
+						if (errno == EINTR &&
+						    !(shf->flags & SHF_INTERRUPT))
+							continue;
+						shf->flags |= SHF_ERROR;
+						shf->errno_ = errno;
+						shf->wnleft = 0;
+						/*
+						 * Note: fwrite(3) returns 0
+						 * for errors - this doesn't
+						 */
+						return (EOF);
+					}
+					buf += n;
+					ncopy -= n;
 				}
-				buf += n;
-				ncopy -= n;
 			}
+			/* ... and buffer the rest */
 		}
 		if (nbytes > 0) {
+			/* write remaining bytes to buffer */
 			memcpy(shf->wp, buf, nbytes);
 			shf->wp += nbytes;
 			shf->wnleft -= nbytes;
 		}
 	}
 
-	return orig_nbytes;
+	return (orig_nbytes);
 }
 
 int
@@ -663,7 +698,7 @@ shf_fprintf(struct shf *shf, const char *fmt, ...)
 	n = shf_vfprintf(shf, fmt, args);
 	va_end(args);
 
-	return n;
+	return (n);
 }
 
 int
@@ -674,14 +709,15 @@ shf_snprintf(char *buf, int bsize, const char *fmt, ...)
 	int n;
 
 	if (!buf || bsize <= 0)
-		internal_errorf("shf_snprintf: buf %p, bsize %d", buf, bsize);
+		internal_errorf("shf_snprintf: buf %lX, bsize %d",
+		    (long)(ptrdiff_t)buf, bsize);
 
 	shf_sopen(buf, bsize, SHF_WR, &shf);
 	va_start(args, fmt);
 	n = shf_vfprintf(&shf, fmt, args);
 	va_end(args);
 	shf_sclose(&shf); /* null terminates */
-	return n;
+	return (n);
 }
 
 char *
@@ -694,7 +730,7 @@ shf_smprintf(const char *fmt, ...)
 	va_start(args, fmt);
 	shf_vfprintf(&shf, fmt, args);
 	va_end(args);
-	return shf_sclose(&shf); /* null terminates */
+	return (shf_sclose(&shf)); /* null terminates */
 }
 
 #undef FP			/* if you want floating point stuff */
@@ -734,7 +770,7 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 	int nwritten = 0;
 
 	if (!fmt)
-		return 0;
+		return (0);
 
 	while ((c = *fmt++)) {
 		if (c != '%') {
@@ -743,11 +779,11 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 			continue;
 		}
 		/*
-		 *  This will accept flags/fields in any order - not
-		 *  just the order specified in printf(3), but this is
-		 *  the way _doprnt() seems to work (on bsd and sysV).
-		 *  The only restriction is that the format character must
-		 *  come last :-).
+		 * This will accept flags/fields in any order - not
+		 * just the order specified in printf(3), but this is
+		 * the way _doprnt() seems to work (on bsd and sysV).
+		 * The only restriction is that the format character must
+		 * come last :-).
 		 */
 		flags = field = precision = 0;
 		for ( ; (c = *fmt++) ; ) {
@@ -820,22 +856,17 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 
 		if (c >= 'A' && c <= 'Z') {
 			flags |= FL_UPPER;
-			c = c - 'A' + 'a';
+			c = ksh_tolower(c);
 		}
 
 		switch (c) {
-		case 'p': /* pointer */
-			flags &= ~(FL_LONG | FL_SHORT);
-			flags |= (sizeof(char *) > sizeof(int)) ?
-			    /* hope it fits.. */ FL_LONG : 0;
-			/* aaahhh... */
 		case 'd':
 		case 'i':
 		case 'o':
 		case 'u':
 		case 'x':
 			flags |= FL_NUMBER;
-			cp = &numbuf[sizeof (numbuf)];
+			cp = numbuf + sizeof(numbuf);
 			/*-
 			 * XXX any better way to do this?
 			 * XXX hopefully the compiler optimises this out
@@ -843,23 +874,23 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 			 * For shorts, we want sign extend for %d but not
 			 * for %[oxu] - on 16 bit machines it doesn't matter.
 			 * Assumes C compiler has converted shorts to ints
-			 * before pushing them.  XXX optimise this -tg
+			 * before pushing them. XXX optimise this -tg
 			 */
 			if (flags & FL_LONG)
 				lnum = va_arg(args, unsigned long);
-			else if ((sizeof (int) < sizeof (long)) && (c == 'd'))
-				lnum = (long) va_arg(args, int);
+			else if ((sizeof(int) < sizeof(long)) && (c == 'd'))
+				lnum = (long)va_arg(args, int);
 			else
 				lnum = va_arg(args, unsigned int);
 			switch (c) {
 			case 'd':
 			case 'i':
-				if (0 > (long) lnum)
-					lnum = - (long) lnum, tmp = 1;
-				else
+				if (0 > (long)lnum) {
+					lnum = -(long)lnum;
+					tmp = 1;
+				} else
 					tmp = 0;
-				/* aaahhhh..... */
-
+				/* FALLTHROUGH */
 			case 'u':
 				do {
 					*--cp = lnum % 10 + '0';
@@ -886,12 +917,9 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 					*--cp = '0';
 				break;
 
-			case 'p':
-			case 'x':
-			    {
+			case 'x': {
 				const char *digits = (flags & FL_UPPER) ?
-				    "0123456789ABCDEF" :
-				    "0123456789abcdef";
+				    digits_uc : digits_lc;
 				do {
 					*--cp = digits[lnum & 0xf];
 					lnum >>= 4;
@@ -901,9 +929,9 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 					*--cp = (flags & FL_UPPER) ? 'X' : 'x';
 					*--cp = '0';
 				}
-			    }
 			}
-			len = &numbuf[sizeof (numbuf)] - (s = cp);
+			}
+			len = numbuf + sizeof(numbuf) - (s = cp);
 			if (flags & FL_DOT) {
 				if (precision > len) {
 					field = precision;
@@ -993,7 +1021,7 @@ shf_vfprintf(struct shf *shf, const char *fmt, va_list args)
 		}
 	}
 
-	return shf_error(shf) ? EOF : nwritten;
+	return (shf_error(shf) ? EOF : nwritten);
 }
 
 #ifdef MKSH_SMALL
